/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <math.h>

#include <rfftw.h>

#include "Context.hpp"
#include "Signal.hpp"
#include "SignalGen.hpp"
#include "RealVector.hpp"

#define INPUT_DIM 64

void calcFreq(RealVector *input, RealVector *ouptut, rfftw_plan plan);

int main(int argc, char *argv[]) {
	char *sigUNL = "/signals/accel";
	char *outputUNL = "/signals/accel_freq";

	Context *context = NULL;
	Signal *signal;
	SignalGen *signalGen;
	RealVector *feature;
	RealVector *inputFeature, *outputFeature;
    rfftw_plan plan;
	int index, i, j;
	float vals[INPUT_DIM];
	int inMiddle;
	unsigned long count;

	printf("AccelFeature\n");
    if(argc >= 2) {
        context = new Context(argv[1]);
    }
    else {
        printf("Usage: AccelFeature [server name]\n");
        exit(0);
    }

	signal = new Signal(sigUNL, context);
	if(signal == NULL) {
		printf("Error allocating signal\n");
		exit(1);
	}
	signalGen = new SignalGen(outputUNL, context);
	if(signalGen == NULL) {
		printf("Error allocating signalGen\n");
		exit(1);
	}

	inputFeature = new RealVector(INPUT_DIM);
	outputFeature = new RealVector(INPUT_DIM/2-1);

    plan = rfftw_create_plan(INPUT_DIM, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

	index = 0;
	inMiddle = 1;

	for(i=0; i<INPUT_DIM; i++) 
		vals[i] = 0;

	count = 0;
	printf("Reading Signals --------\n");
	while(1) {
		signalGen->getConnections(0);
		if( (feature = signal->readReals()) != NULL ) {
			// feature->show();

			vals[index++] = feature->getElement(0);
			if(index == INPUT_DIM/2 || index == INPUT_DIM) {
				if(index == INPUT_DIM)
					index = 0;

				if(inMiddle) {
					j = 0;
					for(i=INPUT_DIM/2; i<INPUT_DIM; i++) 
						inputFeature->setElement(j++, vals[i]);
					for(i=0; i<INPUT_DIM/2; i++) 
						inputFeature->setElement(j++, vals[i]);
				}
				else {
					for(i=0; i<INPUT_DIM; i++) 
						inputFeature->setElement(i, vals[i]);
				}
				inMiddle = ! inMiddle;

				// inputFeature->show();
				calcFreq(inputFeature, outputFeature, plan);
// printf("Freqs *******\n");
// outputFeature->show();

				signalGen->writeReals(outputFeature);
// printf("\n");
			}
			
			if((count % 2000) == 0)
				printf("%ld entries\n", count);
			delete feature;
			count++;
		}
		else {
			printf("Sleeping\n");
			sleep(1);
		}
	}

    rfftw_destroy_plan(plan);

	if(context != NULL) 
		delete context;
	delete signal;
	delete outputFeature;

	return 0;
}

void calcFreq(RealVector *input, RealVector *output, rfftw_plan plan) {
    fftw_real in[INPUT_DIM], out[INPUT_DIM], power_spectrum[INPUT_DIM/2+1];
    int i, j;

	if(input->getNumElements() != INPUT_DIM) {
		printf("Error - input feature vector size does not match expected size %d - %d\n", 
			input->getNumElements(), INPUT_DIM);
		return;
	}

	if(output->getNumElements() != INPUT_DIM/2-1) {
		printf("Error - output feature vector size does not match expected size %d - %d\n", 
			output->getNumElements(), INPUT_DIM/2-1);
		return;
	}

    for(i=0; i<INPUT_DIM; i++) {
        in[i] = input->getElement(i);
    }

    rfftw_one(plan, in, out);

	// for(i=1; i<INPUT_DIM; i++) {
		// printf("out: %.4f\n", out[i]);
	// }

    power_spectrum[0] = out[0]*out[0];  /* DC component */

    for (j = 1; j < (INPUT_DIM+1)/2; ++j)  /* (j < INPUT_DIM/2 rounded up) */
        power_spectrum[j] = out[j]*out[j] + out[INPUT_DIM-j]*out[INPUT_DIM-j];
    if (INPUT_DIM % 2 == 0) /* INPUT_DIM is even */
        power_spectrum[INPUT_DIM/2] = out[INPUT_DIM/2]*out[INPUT_DIM/2];  /* Nyquist freq. */

	// Throw away dc component ( entry 0 )
	for(i=1; i<INPUT_DIM/2; i++) {
        	// printf("%d %f\n", i, power_spectrum[i]);
		output->setElement(i-1, power_spectrum[i]);
	}

}
