/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <math.h>

#include <rfftw.h>

#include "Context.hpp"
#include "Signal.hpp"
#include "SignalGen.hpp"
#include "RealVector.hpp"

#define INPUT_DIM 64
#define OUTPUT_DIM 16

void calcFreq(RealVector *input, RealVector *ouptut, rfftw_plan plan, rfftw_plan plan_inv);

int main(int argc, char *argv[]) {
	char *sigUNL = "/signals/test";
	char *outputUNL = "/signals/cepstrum";

	Context *context = NULL;
	Signal *signal;
	SignalGen *signalGen;
	RealVector *feature;
	RealVector *inputFeature, *outputFeature;
    rfftw_plan plan;
    rfftw_plan plan_inv;
	int index, i, j;
	float vals[INPUT_DIM];
	int inMiddle;

	printf("AudioFeature\n");
    if(argc >= 2) {
        context = new Context(argv[1]);
    }
    else {
        printf("Usage: AudioFeature [server name]\n");
        // exit(0);
    }

	signal = new Signal(sigUNL, context);
	if(signal == NULL) {
		printf("Error allocating signal\n");
		exit(1);
	}
	signalGen = new SignalGen(outputUNL, context);
	if(signalGen == NULL) {
		printf("Error allocating signalGen\n");
		exit(1);
	}

	inputFeature = new RealVector(INPUT_DIM);
	outputFeature = new RealVector(INPUT_DIM/4);

    plan = rfftw_create_plan(INPUT_DIM, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
    plan_inv = rfftw_create_plan(INPUT_DIM/2, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

	index = 0;
	inMiddle = 1;

	for(i=0; i<INPUT_DIM; i++) 
		vals[i] = 0;

	printf("Reading Signals --------\n");
	while(1) {
		signalGen->getConnections(0);
		if( (feature = signal->readReals()) != NULL ) {
			// feature->show();

			vals[index++] = feature->getElement(0);
			if(index == INPUT_DIM/2 || index == INPUT_DIM) {
				if(index == INPUT_DIM)
					index = 0;

				if(inMiddle) {
					j = 0;
					for(i=INPUT_DIM/2; i<INPUT_DIM; i++) 
						inputFeature->setElement(j++, vals[i]);
					for(i=0; i<INPUT_DIM/2; i++) 
						inputFeature->setElement(j++, vals[i]);
				}
				else {
					for(i=0; i<INPUT_DIM; i++) 
						inputFeature->setElement(i, vals[i]);
				}
				inMiddle = ! inMiddle;

				// inputFeature->show();
				calcFreq(inputFeature, outputFeature, plan, plan_inv);
printf("Freqs *******\n");
outputFeature->show();

				signalGen->writeReals(outputFeature);
				printf("\n");
			}
			
			delete feature;
		}
		else {
			printf("Sleeping\n");
			sleep(1);
		}
	}

    rfftw_destroy_plan(plan);
    rfftw_destroy_plan(plan_inv);

	if(context != NULL) 
		delete context;
	delete signal;
	delete outputFeature;

	return 0;
}

void calcFreq(RealVector *input, RealVector *output, rfftw_plan plan, rfftw_plan plan_inv) {
    fftw_real in[INPUT_DIM], out[INPUT_DIM], power_spectrum[INPUT_DIM/2+1];
    int i, j;

	// Error checking
	if(input->getNumElements() != INPUT_DIM) {
		printf("Error - input feature vector size does not match expected size %d - %d\n", 
			input->getNumElements(), INPUT_DIM);
		return;
	}

	if(output->getNumElements() != INPUT_DIM/4) {
		printf("Error - output feature vector size does not match expected size %d - %d\n", 
			output->getNumElements(), INPUT_DIM/4);
		return;
	}

    for(i=0; i<INPUT_DIM; i++) {
        in[i] = input->getElement(i);
    }

	// run FFTW
    rfftw_one(plan, in, out);

	// compute power Spectrum
    power_spectrum[0] = out[0]*out[0];  /* DC component */

    for (j = 1; j < (INPUT_DIM+1)/2; ++j)  /* (j < INPUT_DIM/2 rounded up) */
        power_spectrum[j] = out[j]*out[j] + out[INPUT_DIM-j]*out[INPUT_DIM-j];
    if (INPUT_DIM % 2 == 0) /* INPUT_DIM is even */
        power_spectrum[INPUT_DIM/2] = out[INPUT_DIM/2]*out[INPUT_DIM/2];  /* Nyquist freq. */

	// take the log of the power spectrum
	for(i=0; i<INPUT_DIM/2; i++) {
		in[i] = log(sqrt(power_spectrum[i]));
	}
    rfftw_one(plan_inv, in, out);

	// compute the new power Spectrum
    power_spectrum[0] = out[0]*out[0];
    for (j = 1; j < INPUT_DIM/4; ++j)
        power_spectrum[j] = out[j]*out[j] + out[INPUT_DIM-j]*out[INPUT_DIM-j];
    if (INPUT_DIM % 4 == 0) /* INPUT_DIM is even */
        power_spectrum[INPUT_DIM/4] = out[INPUT_DIM/4]*out[INPUT_DIM/4];  /* Nyquist freq. */

	for(i=0; i<OUTPUT_DIM/4; i++) {
       	// printf("%d %f\n", i, power_spectrum[i]);
		output->setElement(i, power_spectrum[i]);
	}
}
