/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "MarkovList.hpp"
#include "RealVector.hpp"
#include "Matrix.hpp"
#include "Markov.hpp"
#include "Classification.hpp"
#include "IntVector.hpp"

MarkovList::MarkovList() {
}

MarkovList::~MarkovList() {
}

/*************************

File format:

Num Markovs (int)
window size (int)
num elements (int)
Label 1 (string)
prior (double)
init prob (double[])
trans prob (double[][])
Label 2 
.
.
.

**************************/

int MarkovList::getNumElements() {
    return numElements;
}

int MarkovList::getWindowSize() {
    return windowSize;
}

void MarkovList::load(char *filename) {
	FILE *fp;
	char buffer[1024];
	int bufSize = 1023;
	char *ptr;
	int i;

	// printf("Load '%s'\n", filename);

	fp = fopen(filename, "r");
	if(fp == NULL) {
		printf("Error opening file '%s'\n", filename);
		exit(1);
	}

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	numMarkovs = atoi(ptr);

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	windowSize = atoi(ptr);

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	numElements = atoi(ptr);

	markovs = new Markov[numMarkovs];

	for(i=0; i<numMarkovs; i++) {
		readMarkov(fp, markovs+i);
		if(markovs[i].transProb == NULL) {
			printf("Error reading in markovs\n");
			exit(1);
		}

		markovs[i].identifier = i;
	}

	if(0) {
		for(i=0; i<numMarkovs; i++) {
			markovs[i].show();
			printf("\n");
		}
	}

	fclose(fp);
}

void MarkovList::readMarkov(FILE *fp, Markov *markov) {
    char buffer[1024];
    int bufSize = 1023;
    char *ptr, *end;
    int i, j;
	double val;

    markov->numElements = numElements;
    markov->windowSize = windowSize;
	markov->initProb = new RealVector(numElements);
	markov->transProb = new Matrix(numElements, numElements);

	if(markov->initProb == NULL || markov->transProb == NULL) {
		printf("Error allocating Markov structures\n");
		exit(1);
	}

    if(fgets(buffer, bufSize, fp) == NULL)
        return;

    ptr = strchr(buffer, '\n');
    if(ptr != NULL)
        *ptr = '\0';

    ptr = buffer;
    while(*ptr == ' ' || *ptr == '\t')
        ptr++;

    markov->label = strdup(ptr);
    // printf("Label: %s\n", markov->label);

    if(fgets(buffer, bufSize, fp) == NULL)
        return;

    ptr = strchr(buffer, '\n');
    if(ptr != NULL)
        *ptr = '\0';

    ptr = buffer;
    while(*ptr == ' ' || *ptr == '\t')
        ptr++;

    markov->prior = atof(ptr);

    if(fgets(buffer, bufSize, fp) == NULL)
        return;

    ptr = strchr(buffer, '\n');
    if(ptr != NULL)
        *ptr = '\0';

    // printf("u\n");
    ptr = buffer;
    for(i=0; i<numElements; i++) {
        while(*ptr == ' ' || *ptr == '\t')
            ptr++;

        end = ptr;
        while(*end != '\0' && *end != ' ' && *end != '\t')
            end++;
        *end = '\0';

        val = atof(ptr);
        markov->initProb->setElement(i, val);

        ptr = end+1;
    }

    for(i=0; i<numElements; i++) {
        if(fgets(buffer, bufSize, fp) == NULL)
            return;

        ptr = strchr(buffer, '\n');
        if(ptr != NULL)
            *ptr = '\0';

        ptr = buffer;
        for(j=0; j<numElements; j++) {
            while(*ptr == ' ' || *ptr == '\t')
                ptr++;

            end = ptr;
            while(*end != '\0' && *end != ' ' && *end != '\t')
                end++;
            *end = '\0';

            val = atof(ptr);
            markov->transProb->setElement(i, j, val);

            ptr = end+1;
        }
    }

	markov->normalizeProbabilities();
}

//--------------------------------------------------

void MarkovList::classify(Classification *classification, IntVector *states) {
	double prob;
	int index, i;

	classification->zero();
	for(i=0; i<numMarkovs; i++)
		classification->addToLabel(markovs[i].label, markovs[i].calc(states));
// classification->show();
	classification->normalize();

	for(i=0; i<numMarkovs; i++) {
		index = classification->getIndex(markovs[i].label);
		prob = classification->getProbability(index);

		classification->setProbability(index, prob*markovs[i].prior);
	}
}

//--------------------------------------------------

void MarkovList::show() {
	int i;

	for(i=0; i<numMarkovs; i++)
		markovs[i].show();
}

