/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "MixtureModel.hpp"
#include "RealVector.hpp"
#include "Matrix.hpp"
#include "Gaussian.hpp"
#include "Classification.hpp"
#include "RealVector.hpp"

MixtureModel::MixtureModel() {
}

MixtureModel::~MixtureModel() {
}

/*************************

File format:

Num Gaussians (int)
dimensionality (int)
Label 1 (string)
Prior (double)
det(Sigma) (double)
Means (double[])
Sigma^-1 (double[][])
Label 2 
.
.
.

**************************/

void MixtureModel::load(char *filename) {
	FILE *fp;
	char buffer[1024];
	int bufSize = 1023;
	char *ptr;
	int i;

	// printf("Load '%s'\n", filename);

	fp = fopen(filename, "r");
	if(fp == NULL) {
		printf("Error opening file '%s'\n", filename);
		exit(1);
	}

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	numGaussians = atoi(ptr);

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	dimension = atoi(ptr);

	gaussians = new Gaussian[numGaussians];

	for(i=0; i<numGaussians; i++) {
		readGaussian(fp, gaussians+i);
		if(gaussians[i].sigma_inv == NULL) {
			printf("Error reading in gaussians\n");
			exit(1);
		}

		gaussians[i].identifier = i;
	}

	if(0) {
		for(i=0; i<numGaussians; i++) {
			gaussians[i].show();
			printf("\n");
		}
	}

	fclose(fp);

	scaleDets();
}

void MixtureModel::readGaussian(FILE *fp, Gaussian *gaussian) {
	char buffer[1024];
	int bufSize = 1023;
	char *ptr, *end;
	int i, j;
	float val;

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	gaussian->label = strdup(ptr);
	// printf("Label: %s\n", gaussian->label);
	
	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	gaussian->prior = atof(ptr);
	// printf("Prior: %d\n", gaussian->prior);

	gaussian->dimension = dimension;
	// printf("Dimension: %d\n", gaussian->dimension);
	
	gaussian->u = new RealVector(gaussian->dimension);
	gaussian->sigma_inv = new Matrix(gaussian->dimension, gaussian->dimension);

	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	ptr = buffer;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	gaussian->sigma_det = atof(ptr);
	// printf("Sigma_det: %.2f\n", gaussian->sigma_det);
	
	if(fgets(buffer, bufSize, fp) == NULL)
		return;

	ptr = strchr(buffer, '\n');
	if(ptr != NULL)
		*ptr = '\0';

	// printf("u\n");
	ptr = buffer;
	for(i=0; i<gaussian->dimension; i++) {
		while(*ptr == ' ' || *ptr == '\t')
			ptr++;

		end = ptr;
		while(*end != '\0' && *end != ' ' && *end != '\t')
			end++;
		*end = '\0';

		val = atof(ptr);
		gaussian->u->setElement(i, val);

		ptr = end+1;
	}
	// gaussian->u->show();

	// printf("Sigma_inv\n");
	for(i=0; i<gaussian->dimension; i++) {
		if(fgets(buffer, bufSize, fp) == NULL)
			return;

		ptr = strchr(buffer, '\n');
		if(ptr != NULL)
			*ptr = '\0';

		ptr = buffer;
		for(j=0; j<gaussian->dimension; j++) {
			while(*ptr == ' ' || *ptr == '\t')
				ptr++;

			end = ptr;
			while(*end != '\0' && *end != ' ' && *end != '\t')
				end++;
			*end = '\0';

			val = atof(ptr);
			gaussian->sigma_inv->setElement(i, j, val);

			ptr = end+1;
		}
	}
	// gaussian->sigma_inv->show();
}

//--------------------------------------------------

void MixtureModel::show() {
	int i;

	for(i=1; i<numGaussians; i++)
		gaussians[i].show();
}

//--------------------------------------------------

void MixtureModel::scaleDets() {
	double min;
	int i;

	if(numGaussians == 0)
		return;

	min = gaussians[0].sigma_det;

	for(i=1; i<numGaussians; i++) {
		if(min > gaussians[i].sigma_det)
			min = gaussians[i].sigma_det;
	}

	for(i=0; i<numGaussians; i++) 
		gaussians[i].sigma_det /= min;
}

//--------------------------------------------------

void MixtureModel::classify(Classification *classification, RealVector *feature) {
	double prob;
	int index, i;

	classification->zero();
	for(i=0; i<numGaussians; i++)
		classification->addToLabel(gaussians[i].label, gaussians[i].calc(feature));
// classification->show();
	classification->normalize();

	for(i=0; i<numGaussians; i++) {
		index = classification->getIndex(gaussians[i].label);
		prob = classification->getProbability(index);

		classification->setProbability(index, prob*gaussians[i].prior);
	}
}

