/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include "RealVector.hpp"

RealVector::RealVector() {
	numElements = 0;
	element = NULL;
}

RealVector::RealVector(int numVals) {
	int i;

	numElements = numVals;
	element = new double[numElements];
	if(element == NULL) {
		printf("Error allocation %d elements in vector\n", numElements);
		exit(1);
	}

	for(i=0; i<numElements; i++) 
		element[i] = 0;
}

RealVector::~RealVector() {
	if(numElements > 0)
		delete[] element;
}

//--------------------------------------------------

int RealVector::getNumElements() {
	return numElements;
}

double RealVector::getElement(int index) {
	if(index < 0 || index >= numElements)
		return 0;
	else
		return element[index];
}

void RealVector::setElement(int index, double val) {
	if(index < 0 || index >= numElements)
		return;

	element[index] = val;
}

void RealVector::appendElement(double val) {
	double *temp;
	int i;

	temp = new double[numElements+1];
	
	for(i=0; i<numElements; i++)
		temp[i] = element[i];

	if(numElements > 0)
		delete[] element;
	element = temp;
	temp[numElements] = val;
	numElements++;
}

//--------------------------------------------------

void RealVector::show() {
	int i;

	// printf("RealVector::show()\n");
    // printf("  numElements: %d\n", numElements);

	printf("Vect: ");
	for(i=0; i<numElements; i++) {
		printf("%e ", element[i]);
		// printf("%.2f ", element[i]);
	}
	printf("\n");
}

//--------------------------------------------------

void RealVector::add(RealVector *v1, double num, RealVector *result) {
	int i;

	for(i=0; i<v1->numElements; i++)
		result->element[i] = v1->element[i]+num;
}

void RealVector::add(RealVector *v1, RealVector *v2, RealVector *result) {
	int i;

	if(v1->numElements != v2->numElements) {
		printf("Error - add vectors of different lengths\n");
	}

	for(i=0; i<v1->numElements; i++)
		result->element[i] = v1->element[i]+v2->element[i];
}

void RealVector::subtract(RealVector *v1, double num, RealVector *result) {
	int i;

	for(i=0; i<v1->numElements; i++)
		result->element[i] = v1->element[i]-num;
}

void RealVector::subtract(RealVector *v1, RealVector *v2, RealVector *result) {
	int i;

	if(v1->numElements != v2->numElements) {
		printf("Error - subtract vectors of different lengths\n");
	}

	for(i=0; i<v1->numElements; i++)
		result->element[i] = v1->element[i]-v2->element[i];
}

void RealVector::multiply(RealVector *v1, double num, RealVector *result) {
	int i;

	for(i=0; i<v1->numElements; i++)
		result->element[i] = v1->element[i]*num;
}

//--------------------------------------------------

double RealVector::innerProduct(RealVector *v1, RealVector *v2) {
	int i;
	double sum = 0;

	if(v1->numElements != v2->numElements) {
		printf("Error - inner product vectors of different lengths\n");
		return 0;
	}

	for(i=0; i<v1->numElements; i++)
		sum += v1->element[i]*v2->element[i];

	return sum;
}

void RealVector::outerProduct(RealVector *v1, RealVector *v2, Matrix *result) {
	int i, j;

	for(i=0; i<v1->numElements; i++) {
		for(j=0; j<v1->numElements; j++) {
			result->setElement(i, j, v1->element[i]*v2->element[j]);			
		}
	}
}

void RealVector::multiply(RealVector *v1, Matrix *m1, RealVector *result) {
    int i, j;
    double sum;

    if(v1->numElements != m1->getNumRows()) {
        printf("Error - mismatched vector and matrix sizes for multiply\n");
        return;
    }

    if(m1->getNumCols() != result->numElements) {
        printf("Error - mismatched vector and matrix sizes for multiply\n");
        return;
    }

	for(i=0; i<m1->getNumCols(); i++) {
		sum = 0;
		for(j=0; j<m1->getNumRows(); j++) {
			sum += v1->element[j]*m1->getElement(j,i);
			// printf("v1[%d] * m1[%d,%d] = %.2f * %.2f = %.2f\n", 
				// j, j, i, v1->element[j], m1->getElement(j,i), v1->element[j]*m1->getElement(j,i));
		}
		result->setElement(i, sum);			
	}
}
