/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "Signal.hpp"
#include "RealVector.hpp"
#include "IntVector.hpp"
#include "Context.hpp"

Signal::Signal() {
	unl = NULL;
	socket_id = -1;
	
	port = 0;
	ip = 0;

	context = NULL;

	openConnection();

	if( socket_id == -1 ) {
        printf("Connection to socket failed\n");
        // exit(1);
    }
}

Signal::Signal(char *newUNL, Context *newContext) {
	unl = strdup(newUNL);
	socket_id = -1;
	
	port = 0;
	ip = 0;

	context = newContext;
	openConnection();

	if( socket_id == -1 ) {
        printf("Connection to socket failed\n");
        // exit(1);
    }
}

Signal::~Signal() {
	if(socket_id != -1)
		close(socket_id);

	free(unl);
}

void Signal::setContext(Context *newContext) {
	context = newContext;
}

void Signal::setUNL(char *newUNL) {
	if(unl != NULL)
		free(unl);
	unl = strdup(newUNL);
}

void Signal::openConnection() {
    struct sockaddr_in addr;

    socket_id = socket(AF_INET, SOCK_STREAM, 0);
    if(socket_id == -1) {
        printf("Error creating socket\n");
        exit(1);
    }

	if(! getSignalAddress(&addr) ) {
		// printf("Error getting Address\n");
		return;
	}

	if(connect(socket_id, (struct sockaddr*) &addr, sizeof(struct sockaddr_in)) == -1) {
        // printf("Error connecting to socket\n");
		socket_id = -1;
        return;
    }
}

RealVector* Signal::readReals() {
	int i, numBytes;
	long numEntries;
	unsigned long start_word;
	float val;
	RealVector *returnVect = NULL;
	unsigned long buffer;

	if(socket_id == -1)
		openConnection();

	if(socket_id != -1) {
		// printf("Read Signal\n");
		
		numBytes = read(socket_id, &buffer, 4);
		if(numBytes != 4) {
			if(numBytes <= 0)
				closeConnection();
			return NULL;
		}
		start_word = ntohl(buffer);
		if(start_word != START_VAL) {
			printf("Error - invalid start of record 0x%lx\n", start_word);
			return NULL;
		}
		// printf("start_word: 0x%lx\n", start_word);

		numBytes = read(socket_id, &buffer, 4);
		if(numBytes != 4) {
			if(numBytes <= 0)
				closeConnection();
			return NULL;
		}
		numEntries = ntohl(buffer);
		// printf("NumEntries: %ld\n", numEntries);

		returnVect = new RealVector(numEntries);

		for(i=0; i < numEntries; i++) {
			numBytes = read(socket_id, &buffer, 4);
			if(numBytes != 4) {
				if(numBytes <= 0)
					closeConnection();
				printf("Error reading vector %d\n", numBytes);
				return returnVect;
			}
			// printf(" Val: %.2f\n", val);
			buffer = ntohl(buffer);
			val = *((float*) &buffer);
			returnVect->setElement(i, val);
		}
	}

	return returnVect;
}

IntVector* Signal::readInts() {
	int i, numBytes;
	long numEntries;
	unsigned long start_word;
	unsigned long val;
	IntVector *returnVect = NULL;
	unsigned long buffer;

	if(socket_id == -1)
		openConnection();

	if(socket_id != -1) {
		// printf("Read Signal\n");
		
		numBytes = read(socket_id, &buffer, 4);
		if(numBytes != 4) {
			if(numBytes <= 0)
				closeConnection();
			return NULL;
		}
		start_word = ntohl(buffer);
		if(start_word != START_VAL) {
			printf("Error - invalid start of record 0x%lx\n", start_word);
			return NULL;
		}
		// printf("start_word: 0x%lx\n", start_word);

		numBytes = read(socket_id, &buffer, 4);
		if(numBytes != 4) {
			if(numBytes <= 0)
				closeConnection();
			return NULL;
		}
		numEntries = ntohl(buffer);
		// printf("NumEntries: %ld\n", numEntries);

		returnVect = new IntVector(numEntries);

		for(i=0; i < numEntries; i++) {
			numBytes = read(socket_id, &buffer, 4);
			if(numBytes != 4) {
				if(numBytes <= 0)
					closeConnection();
				printf("Error reading vector %d\n", numBytes);
				return returnVect;
			}
			// printf(" Val: %d\n", val);
			buffer = ntohl(buffer);
			val = buffer;
			returnVect->setElement(i, val);
		}
	}

	return returnVect;
}

void Signal::closeConnection() {
	if(socket_id != -1) {
printf("Closing Socket\n");
		close(socket_id);
		socket_id = -1;
	}
}

int Signal::getSignalAddress(struct sockaddr_in *addr) {
	NodeType *topNode;
	NodeType *ipNode;
	NodeType *portNode;
	char *ipStr = InstanceString("ip");
	char *portStr = InstanceString("port");
	
	if(context == NULL) {
		printf("Error - getSignalAddress with no context\n");
		exit(1);
	}

	if(unl == NULL) {
		printf("Error - getSignalAddress with no UNL\n");
		exit(1);
	}

	topNode = context->getNode(unl);
	if(topNode == NULL) {
		printf("Cannot get node '%s'\n", unl);
		return 0;
	}

	ipNode = GetChild(topNode, ipStr);
	portNode = GetChild(topNode, portStr);

	if(ipNode == NULL || portNode == NULL) {
		printf("Error getting ip and port numbers\n");
		return 0;
	}

	port = portNode->data.long_val;
	ip = ipNode->data.long_val;

	DeleteNode(NULL, topNode);

    bzero(addr, sizeof(struct sockaddr_in));
    addr->sin_family = AF_INET;
    addr->sin_port = htons(port);
	addr->sin_addr.s_addr = htonl(ip);

	// printf("getSignalAddress - got node - ip: 0x%lx - '%s' port: %d\n", ip, inet_ntoa(addr->sin_addr), port);

	return 1;
}

