/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/poll.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "SignalGen.hpp"
#include "Context.hpp"

SignalGen::SignalGen() {
	numListeners = 0;
	listeners = NULL;

	context = NULL;

	signal(SIGPIPE, SIG_IGN);

	baseSocket = 0;
	unl = NULL;

	ip = 0;
	port = 0;
}

SignalGen::SignalGen(char *newUNL, Context *newContext) {
	numListeners = 0;
	listeners = NULL;

	context = newContext;

	signal(SIGPIPE, SIG_IGN);

	baseSocket = 0;
	unl = NULL;

	ip = 0;
	port = 0;

	startListener(newUNL, newContext);
}

SignalGen::~SignalGen() {
	int i;

	for(i=0; i<numListeners; i++) 
		close(listeners[i]);
	if(listeners != NULL) 
		delete[] listeners;

	if(baseSocket != 0) {
		printf("Closing baseSocket\n");
		close(baseSocket);
	}
	if(unl != NULL)
		free(unl);
}

void SignalGen::startListener(char *newUNL, Context *newContext) {
	struct sockaddr_in addr;
	int MAX_BACKLOG = 10;

	if(unl != NULL)
		free(unl);

	unl = strdup(newUNL);
	printf("Connecting SignalGen for \"%s\"\n", newUNL);

	context = newContext;

	if(baseSocket != 0)
		close(baseSocket);

	baseSocket = socket(AF_INET, SOCK_STREAM, 0);
    if(baseSocket == -1) {
        printf("Error creating socket\n");
        exit(1);
    }

    bzero(&addr, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(0);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);

    if(bind(baseSocket, (struct sockaddr*) &addr, sizeof(struct sockaddr_in)) < 0) {
        printf("Error binding to socket\n");
        exit(1);
    }

    if(listen(baseSocket, MAX_BACKLOG) == -1) {
        printf("Error - cannot listen on specified TCP port\n");
        exit(1);
    }

	announceAddress();
}

void SignalGen::getConnections(int waitMillisec) {
	int sock;
	socklen_t incomingAddrSize;
	struct sockaddr_in incomingAddr;
	struct pollfd poll_connect[1];

	if(baseSocket == 0) {
		printf("Error - attempted to getConnections before startListener has been called\n");
		return;
	}

	poll_connect[0].fd = baseSocket;
	poll_connect[0].events = POLLIN;
	poll_connect[0].revents = 0;

	if( poll(poll_connect, 1, waitMillisec)  && (poll_connect[0].revents & POLLIN)) {
		// printf("Poll 0x%x\n", poll_connect[0].revents);
		incomingAddrSize = sizeof(struct sockaddr_in);
		sock = accept(baseSocket, (struct sockaddr*) &incomingAddr, &incomingAddrSize);
		if(sock == -1) {
			perror("accept socket");
			printf("Error accepting socket\n");
			exit(1);
		}
printf("Accepted socket %d\n", sock);
		addListener(sock);
	}
}

void SignalGen::writeReals(RealVector *feature) {
    int i, j;
	int numElements;
	float element;
	unsigned long buffer;

	// printf("SignalGen::write()\n");

	if(feature == NULL) {
		printf("Empty feature\n");
		return;
	}

	numElements = feature->getNumElements();

	for(j=0; j<numListeners; j++) {
// printf("Write to Listener %d socket %d\n", j, listeners[j]);
		buffer = htonl(START_VAL);
		if(write(listeners[j], (const void*) &buffer, 4) != 4) {
			killConnection(listeners[j]);
			return;
		}
			
		buffer = htonl(numElements);
		if(write(listeners[j], (const void*) &buffer, 4) != 4) {
			killConnection(listeners[j]);
			return;
		}

		for(i=0; i<numElements; i++) {
			element = feature->getElement(i);
// printf("Writing val: %.2f\n", element);
			buffer = htonl( *((unsigned long*) &element));
			if(write(listeners[j], (const void*) &buffer, 4) != 4) {
				killConnection(listeners[j]);
				return;
			}
		}
	}
// printf("\n");
}

void SignalGen::writeInts(IntVector *feature) {
    int i, j;
	int numElements;
	unsigned long element;
	unsigned long buffer;

	// printf("SignalGen::write()\n");

	if(feature == NULL) {
		printf("Empty feature\n");
		return;
	}

	numElements = feature->getNumElements();

	for(j=0; j<numListeners; j++) {
// printf("Write to Listener %d socket %d\n", j, listeners[j]);
		buffer = htonl(START_VAL);
		if(write(listeners[j], (const void*) &buffer, 4) != 4) {
			killConnection(listeners[j]);
			return;
		}
			
		buffer = htonl(numElements);
		if(write(listeners[j], (const void*) &buffer, 4) != 4) {
			killConnection(listeners[j]);
			return;
		}

		for(i=0; i<numElements; i++) {
			element = feature->getElement(i);
// printf("Writing val: %.2f\n", element);
			buffer = htonl( element );
			if(write(listeners[j], (const void*) &buffer, 4) != 4) {
				killConnection(listeners[j]);
				return;
			}
		}
	}
// printf("\n");
}

void SignalGen::addListener(int newSocket) {
	int *temp;
	int i;

	temp = new int[numListeners+1];
	
	for(i=0; i<numListeners; i++) {
		temp[i] = listeners[i];
	}
	delete[] listeners;

	listeners = temp;
	listeners[numListeners] = newSocket;

	numListeners++;
}

void SignalGen::removeListener(int newSocket) {
	int i;

	for(i=0; i<numListeners; i++) {
		if(listeners[i] == newSocket)
			break;
	}

	if(i != numListeners) {
		printf("Killing socket %d\n", newSocket);
		for(; i<numListeners-1; i++)
			listeners[i] = listeners[i+1];
		numListeners--;
	}
}

void SignalGen::announceAddress() {
    NodeType *updateNode;
    NodeType *ipNode;
    NodeType *portNode;
    char *nodeName;
	struct sockaddr_in addr;
	socklen_t addrLen;
	struct hostent *hent;

    printf("SignalGen::announceAddress to %s\n", unl);

    if(context == NULL)
        return;

	if(baseSocket == 0) {
		printf("Error announceAddress on empty socket\n");
		exit(1);
	}

	addrLen = sizeof(struct sockaddr_in);
	if( getsockname(baseSocket, (struct sockaddr*) &addr, &addrLen) != 0) {
		perror("getsockname");
		exit(1);
	}

	char buffer[256];
	if(gethostname(buffer, sizeof(buffer)) == 0)
		printf("Hostname: '%s'\n", buffer);
	else
		printf("Could not get hostname\n");

	hent = gethostbyname(buffer);
	if(hent == NULL) {
		printf("Error getting host by name\n");
		exit(1);
	}
	else {
		int i = 0;

		while(hent->h_addr_list[i] != NULL) {
			printf("Host: %s\n", inet_ntoa( *(struct in_addr *) hent->h_addr_list[i]));
			i++;
		}
	}

	addr.sin_addr =  *((struct in_addr *) hent->h_addr_list[0]);
	ip = ntohl(addr.sin_addr.s_addr);
	port = ntohs(addr.sin_port);

	printf("  Address: '%s' Port: %d\n", inet_ntoa(addr.sin_addr), port);

    nodeName = strrchr(unl, '/');
    if(nodeName == NULL) {
        printf("Error extracting node name from partial unl\n");
        return;
    }
    nodeName++;

    updateNode = CreateBranchNode(nodeName, T_BRANCH, 1);

	ipNode = CreateNode("ip", T_LONG, 4, NULL, FALSE);
	InsertNode(updateNode, ipNode);
	ipNode->data.long_val = ip;

	portNode = CreateNode("port", T_LONG, 4, NULL, FALSE);
	InsertNode(updateNode, portNode);
	portNode->data.long_val = port;

    context->publish(updateNode, unl);

    // PrintNode(updateNode, NULL);
    DeleteNode(NULL, updateNode);
}

void SignalGen::killConnection(int sock) {
	printf("Client lost\n");
	close(sock);
	removeListener(sock);
}

int SignalGen::hasListeners() {
	return (numListeners > 0);
}

