/*****************************************************************************
 *
 * Author: Steve Dunn, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <rfftw.h>
#include "Vector.hpp"
#include "Matrix.hpp"
#include "RealVector.hpp"
#include "Classification.hpp"
#include "MixtureModel.hpp"

#define N 64

#define FULL_CALC 1
#define VERBOSE_OUTPUT 0

int main(int argc, char *argv[]) {
    fftw_real in[N], out[N], power_spectrum[N/2+1];
	rfftw_plan p;
	int i, j;
	FILE *fp;
	char buffer[256];
	int setNum;
	RealVector *feature;
	Classification *classification;
	MixtureModel mixtureModel;
	char *dataFile;
	int index;

	if(argc < 3) {
		printf("usage: fftTest [filename] [model]\n");
		exit(0);
	}

	dataFile = argv[2];

	fp = fopen(argv[1], "r");
	if(fp == NULL) {
		perror("fopen");
		exit(1);
	}

    p = rfftw_create_plan(N, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

	for(i=0; i<N; i++) 
		in[i] = 0;

	if(FULL_CALC) {
		feature = new RealVector(N/2-1);
		classification = new Classification();
		mixtureModel.load(dataFile);

		if(VERBOSE_OUTPUT) {
			for(i=0; i<mixtureModel.numGaussians; i++)
				mixtureModel.gaussians[i]._Verbose = 1;
		}
	}

	setNum = 0;
	while(1) {
		for(i=0; i<N/2; i++) 
			in[i] = in[i+N/2];

		for(i=0; i<N/2; i++) {
			if(fgets(buffer, sizeof(buffer), fp) == NULL) {

				rfftw_destroy_plan(p);
				fclose(fp);
				exit(0);
			}

			in[i+N/2] = atof(buffer);
		}
		
		if(setNum == 0) {
			setNum++;
			continue;
		}

		if(VERBOSE_OUTPUT)
			printf("Set: %d\n", setNum-1);

		rfftw_one(p, in, out);

		// for(i=0; i<N/2; i++) 
			// printf("%11.4f + %11.4fi\n", out[i], out[N-i]);

		power_spectrum[0] = out[0]*out[0];  /* DC component */

		for (j = 1; j < (N+1)/2; ++j)  /* (j < N/2 rounded up) */
			power_spectrum[j] = out[j]*out[j] + out[N-j]*out[N-j];
		if (N % 2 == 0) /* N is even */
			power_spectrum[N/2] = out[N/2]*out[N/2];  /* Nyquist freq. */

		if(! FULL_CALC) {
			for(i=1; i<N/2-1; i++) {
				// power_spectrum[i] = sqrt(power_spectrum[i]);
				printf("%.4f ", power_spectrum[i]);
			}
			// power_spectrum[i] = sqrt(power_spectrum[i]);
			printf("%.4f\n", power_spectrum[i]);
		}
		else {
			for(i=1; i<N/2; i++) {
				feature->setElement(i-1, (power_spectrum[i]));
				// feature->setElement(i-1, sqrt(power_spectrum[i]));
			}
			if(VERBOSE_OUTPUT) {
				printf("PowerSpectrum:\n");
				feature->show();
				printf("Calc:\n");
			}
			mixtureModel.classify(classification, feature);
			if(VERBOSE_OUTPUT) {
				printf("Normalized:\n");
				classification->show();
				printf("\n");
			}
			else {
				index = classification->mostLikelyIndex();
				if(index == -1)
					printf("%4d None 0.0\n", setNum);
				else
					// printf("%4d %s %.4f %d\n", setNum, classification->getLabel(index), classification->getProbability(index), index+1);
					printf("%d\n", index+1);
			}
			
			setNum++;
		}
	}

	rfftw_destroy_plan(p);
	fclose(fp);

	return 0;
}

